import os
import io
import logging
import requests
from queue import Queue
from threading import Event

import google.auth
from google.cloud import firestore
from google.cloud.firestore import DocumentReference
from google.api_core.retry import Retry, if_exception_type
from google.api_core.exceptions import Unknown
from google.auth.exceptions import DefaultCredentialsError

from burla._auth import AuthException, get_gcs_credentials

# throws some uncatchable, unimportant, warnings
logging.getLogger("google.api_core.bidi").setLevel(logging.ERROR)


class InputTooBig(Exception):
    pass


class UnknownClusterError(Exception):
    def __init__(self):
        msg = "An unknown error occurred inside your Burla cluster, "
        msg += "this is not an error with your code."
        super().__init__(msg)


def get_host():
    # not defined in init because users often change this post-import
    custom_host = os.environ.get("BURLA_API_URL")
    return custom_host or "https://cluster.burla.dev"


def using_demo_cluster():
    # not defined in init because users often change this post-import
    return not bool(os.environ.get("BURLA_API_URL"))


def get_db(auth_headers: dict):
    if using_demo_cluster():
        credentials = get_gcs_credentials(auth_headers)
        return firestore.Client(credentials=credentials, project="burla-prod")
    else:
        # use user's local google project/creds if not using our cluster.
        try:
            credentials, project = google.auth.default()
            return firestore.Client(credentials=credentials, project=project)
        except DefaultCredentialsError as e:
            raise Exception(
                "No Google Application Default Credentials found. "
                "Please ensure you have a valid Google Cloud account and are logged in."
            ) from e


def healthcheck_job(job_id: str, auth_headers: dict):
    response = requests.get(f"{get_host()}/v1/jobs/{job_id}", headers=auth_headers)
    if response.status_code == 401:
        raise AuthException()
    elif response.status_code == 404:
        # this thread often runs for a bit after the job has ended, causing 404s
        # for now, just ignore these.
        pass
    else:
        UnknownClusterError()


def print_logs_from_db(
    job_doc_ref: DocumentReference, stop_event: Event, log_msg_stdout: io.TextIOWrapper
):

    def on_snapshot(collection_snapshot, changes, read_time):
        for change in changes:
            if change.type.name == "ADDED":
                log_msg_stdout.write(change.document.to_dict()["msg"])

    collection_ref = job_doc_ref.collection("logs")
    query_watch = collection_ref.on_snapshot(on_snapshot)

    while not stop_event.is_set():
        stop_event.wait(0.5)  # this does not block the processing of new documents
    query_watch.unsubscribe()


def enqueue_results_from_db(job_doc_ref: DocumentReference, stop_event: Event, queue: Queue):
    def on_snapshot(collection_snapshot, changes, read_time):
        for change in changes:
            if change.type.name == "ADDED":
                result = change.document.to_dict()
                result_tuple = (change.document.id, result["is_error"], result["result_pkl"])
                queue.put(result_tuple)

    collection_ref = job_doc_ref.collection("results")
    query_watch = collection_ref.on_snapshot(on_snapshot)

    while not stop_event.is_set():
        stop_event.wait(0.5)  # this does not block the processing of new documents
    query_watch.unsubscribe()


def upload_inputs(DB: firestore.Client, inputs_id: str, inputs_pkl: list[bytes], stop_event: Event):
    """
    Uploads inputs into a separate collection not connected to the job
    so that uploading can start before the job document is created.
    """
    batch_size = 100
    inputs_parent_doc = DB.collection("inputs").document(inputs_id)

    firestore_commit_retry_policy = Retry(
        initial=5.0,
        maximum=120.0,
        multiplier=2.0,
        deadline=900.0,
        predicate=if_exception_type(Unknown),
        reraise=True,
    )

    total_n_bytes_firestore_batch = 0
    firestore_batch = DB.batch()

    for batch_min_index in range(0, len(inputs_pkl), batch_size):
        batch_max_index = batch_min_index + batch_size
        input_batch = inputs_pkl[batch_min_index:batch_max_index]
        subcollection = inputs_parent_doc.collection(f"{batch_min_index}-{batch_max_index}")

        for local_input_index, input_pkl in enumerate(input_batch):
            input_index = local_input_index + batch_min_index
            input_too_big = len(input_pkl) > 1_000_000  # 1MB size limit per firestore doc

            if stop_event.is_set():
                return

            # if batch will contain too much data (10MB), push it before adding input to next batch.
            if total_n_bytes_firestore_batch + len(input_pkl) > 10_000_000:
                firestore_batch.commit(retry=firestore_commit_retry_policy)
                firestore_batch = DB.batch()
                total_n_bytes_firestore_batch = 0

            if input_too_big:
                msg = f"Input at index {input_index} is greater than 1MB in size.\n"
                msg += "Individual inputs greater than 1MB in size are currently not supported."
                raise InputTooBig(msg)
            else:
                doc_ref = subcollection.document(str(input_index))
                firestore_batch.set(doc_ref, {"input": input_pkl, "claimed": False})
                total_n_bytes_firestore_batch += len(input_pkl)

    firestore_batch.commit(retry=firestore_commit_retry_policy)



import os
from fire import Fire


# needed so main_service can associate a client version with a request
__version__ = "0.9.5"
_BURLA_BACKEND_URL = "https://backend.burla.dev"

from burla._auth import login
from burla._remote_parallel_map import remote_parallel_map


def init_cli():
    Fire({"login": login, "version": lambda: print(__version__)})



import io
import sys
import requests
import pickle
import json
from six import reraise
from threading import Thread, Event
from typing import Callable, Optional
from time import sleep
from queue import Queue
from uuid import uuid4
from requests import Response

import cloudpickle
from google.cloud import firestore
from yaspin import yaspin
from tblib import Traceback

from burla import __version__
from burla._auth import get_auth_headers, AuthException
from burla._helpers import (
    upload_inputs,
    print_logs_from_db,
    enqueue_results_from_db,
    healthcheck_job,
    get_db,
    get_host,
)

MAX_PARALLELISM = 1000  # outdated.

# This MUST be set to the same value as `JOB_HEALTHCHECK_FREQUENCY_SEC` in the node service.
# Nodes will restart themself if they dont get a new healthcheck from the client every X seconds.
JOB_HEALTHCHECK_FREQUENCY_SEC = 3


class MainServiceError(Exception):
    # Error from inside the main_service that should be passed back to the user.
    def __init__(self, response: Response):
        self.__class__.__name__ = response.json().get("error_type")
        super().__init__(response.json().get("message"))


class InputsTooBig(Exception):
    pass


def _start_job(
    function_: Callable,
    inputs: list,
    func_cpu: int,
    func_ram: int,
    max_parallelism: int,
    stop_event: Event,
    db: firestore.Client,
    auth_headers: dict,
) -> str:

    inputs_pkl = [cloudpickle.dumps(input_) for input_ in inputs]
    inputs_size = sum([len(input_pkl) for input_pkl in inputs_pkl])
    if inputs_size > 84_866_368:
        raise Exception("Total size of all inputs exceeds current maximum limit of 84.8MB")

    # in separate thread start uploading inputs:
    inputs_id = str(uuid4())
    args = (db, inputs_id, inputs_pkl, stop_event)
    input_uploader_thread = Thread(target=upload_inputs, args=args, daemon=True)
    input_uploader_thread.start()

    payload = {
        "n_inputs": len(inputs),
        "inputs_id": inputs_id,
        "func_cpu": func_cpu,
        "func_ram": func_ram,
        "max_parallelism": max_parallelism,
        "python_version": f"3.{sys.version_info.minor}",
        "burla_version": __version__,
    }

    # tell service to start job
    url = f"{get_host()}/v1/jobs/"
    files = {"function_pkl": cloudpickle.dumps(function_)}
    data = dict(request_json=json.dumps(payload))

    response = requests.post(url, files=files, data=data, headers=auth_headers)
    response_is_json = response.headers.get("Content-Type") == "application/json"

    if response.status_code == 401:
        stop_event.set()
        raise AuthException()
    elif response.status_code != 200 and response_is_json and response.json().get("error_type"):
        stop_event.set()
        raise MainServiceError(response)
    else:
        response.raise_for_status()
        return response.json()["job_id"]


def _watch_job(
    job_id: str,
    n_inputs: int,
    log_msg_stdout: io.TextIOWrapper,
    stop_event: Event,
    db: firestore.Client,
    auth_headers: dict,
):
    job_doc_ref = db.collection("jobs").document(job_id)

    # Start printing logs generated by this job from a separate thread.
    args = (job_doc_ref, stop_event, log_msg_stdout)
    log_thread = Thread(target=print_logs_from_db, args=args, daemon=True)
    log_thread.start()

    # Start enqueueing results (either return-values or errors) generated by this job
    # from a separate thread.
    result_queue = Queue()
    args = (job_doc_ref, stop_event, result_queue)
    result_thread = Thread(target=enqueue_results_from_db, args=args, daemon=True)
    result_thread.start()

    time_since_last_healthcheck = 0
    n_results_received = 0
    while n_results_received < n_inputs:
        sleep(0.05)
        time_since_last_healthcheck += 0.05

        if time_since_last_healthcheck > JOB_HEALTHCHECK_FREQUENCY_SEC:
            healthcheck_job(job_id=job_id, auth_headers=auth_headers)
            time_since_last_healthcheck = 0

        while not result_queue.empty():
            n_results_received += 1
            input_index, is_error, result_pkl = result_queue.get()

            if is_error:
                exc_info = pickle.loads(result_pkl)
                traceback = Traceback.from_dict(exc_info["traceback_dict"]).as_traceback()
                reraise(tp=exc_info["type"], value=exc_info["exception"], tb=traceback)
            else:
                yield cloudpickle.loads(result_pkl)

    stop_event.set()


def remote_parallel_map(
    function_: Callable,
    inputs: list,
    func_cpu: int = 1,
    func_ram: int = 4,
    spinner: bool = True,
    generator: bool = False,
    max_parallelism: Optional[int] = None,
    api_key: Optional[str] = None,
):
    """
    Run an arbitrary Python function on many remote computers in parallel.

    Run provided function_ on each item in inputs at the same time, each on a separate CPU,
    up to 256 CPUs (as of 1/3/25). If more than 256 inputs are provided, inputs are queued and
    processed sequentially on each worker. Any exception raised by `function_`
    (including its stack trace) will be re-raised on the client machine.

    Args:
        function_ (Callable):
            A Python function that accepts a single input argument. For example, calling
            `function_(inputs[0])` should not raise an exception.
        inputs (Iterable[Any]):
            An iterable of elements that will be passed to `function_`.
        func_cpu (int, optional):
            The number of CPUs allocated for each instance of `function_`. The maximum allowable
            value is 32. Defaults to 1.
        func_ram (int, optional):
            The amount of RAM (in GB) allocated for each instance of `function_`. The maximum
            allowable value is 128. Defaults to 4.
        spinner (bool, optional):
            If set to False, disables the display of the status indicator/spinner. Defaults to True.
        generator (bool, optional):
            If True, returns a generator that yields outputs as they are produced; otherwise,
            returns a list of outputs once all have been processed. Defaults to False.
        max_parallelism (int, optional):
            The maximum number of `function_` instances allowed to be running at the same time.
            Defaults to the number of available CPUs divided by `func_cpu`.
        api_key (str, optional):
            An API key for use in deployment environments where `burla login` cannot be run.

    Returns:
        List[Any] or Generator[Any, None, None]:
            A list containing the objects returned by `function_` in no particular order.
            If `generator=True`, returns a generator that yields results as they are produced.

    See Also:
        For more info see our overview: https://docs.burla.dev/overview
        or API-Reference: https://docs.burla.dev/api-reference
    """
    max_parallelism = max_parallelism if max_parallelism else len(inputs)
    max_parallelism = max_parallelism if max_parallelism < MAX_PARALLELISM else MAX_PARALLELISM
    if spinner:
        spinner = yaspin()
        spinner.text = f"Preparing to run {len(inputs)} inputs through `{function_.__name__}`"
        spinner.start()

    auth_headers = get_auth_headers(api_key)
    db = get_db(auth_headers)

    # wrap user function with a for loop because sending too many inputs causes firestore issues
    # this is a temporary fix:
    max_inputs = min(len(inputs), 256)
    batch_size = len(inputs) // max_inputs
    remainder = len(inputs) % max_inputs
    start = 0
    input_batches = []
    for i in range(max_inputs):
        end = start + batch_size + (1 if i < remainder else 0)
        input_batches.append(inputs[start:end])
        start = end

    def function_wrapped(input_batch):
        return [function_(input_) for input_ in input_batch]

    #
    #

    stop_event = Event()
    try:
        job_id = _start_job(
            function_=function_wrapped,
            inputs=input_batches,
            func_cpu=func_cpu,
            func_ram=func_ram,
            max_parallelism=max_parallelism,
            stop_event=stop_event,
            db=db,
            auth_headers=auth_headers,
        )
        log_msg_stdout = spinner if spinner else sys.stdout
        output_batch_generator = _watch_job(
            job_id=job_id,
            n_inputs=len(input_batches),
            log_msg_stdout=log_msg_stdout,
            stop_event=stop_event,
            db=db,
            auth_headers=auth_headers,
        )

        if spinner:
            spinner.text = f"Running {len(inputs)} inputs through `{function_.__name__}`"

        def _output_generator():
            # yield from output_batch_generator
            for output_batch in output_batch_generator:
                yield from output_batch

        if not generator:
            results = [item for item in _output_generator()]

    except Exception as e:
        if spinner:
            spinner.stop()
        raise e
    finally:
        stop_event.set()

    if spinner:
        spinner.text = "Done!"
        spinner.ok("✔")

    if generator:
        return _output_generator()
    else:
        return results



import os
import json
import webbrowser
import requests
from time import sleep
from uuid import uuid4
from pathlib import Path
from typing import Tuple, Optional

from google.oauth2 import service_account
from appdirs import user_config_dir

from burla import _BURLA_BACKEND_URL

AUTH_TIMEOUT_SECONDS = 180
BURLA_APPDATA_DIR = Path(user_config_dir(appname="burla", appauthor="burla"))
CONFIG_PATH = BURLA_APPDATA_DIR / Path("burla_credentials.json")
IN_COLAB = os.getenv("COLAB_RELEASE_TAG") is not None


class AuthTimeoutException(Exception):
    def __init__(self):
        super().__init__("Timed out waiting for authentication flow to complete.")


class AuthException(Exception):
    def __init__(self):
        super().__init__(
            "You are not logged in! Please run `burla login` to create an account or login."
        )


def get_gcs_credentials(burla_auth_headers: dict):
    url = f"{_BURLA_BACKEND_URL}/v1/private/svc_account"
    response = requests.get(url, headers=burla_auth_headers)
    response.raise_for_status()
    service_account_info = json.loads(response.json())
    credentials = service_account.Credentials.from_service_account_info(service_account_info)
    return credentials


def get_auth_headers(api_key: Optional[str] = None) -> Tuple[str, str]:
    login_credentials_missing = not CONFIG_PATH.exists()

    if api_key:
        return {"Authorization": f"Bearer {api_key}"}
    elif login_credentials_missing:
        raise AuthException()
    else:
        auth_info = json.loads(CONFIG_PATH.read_text())
        return {"email": auth_info["email"], "Authorization": f"Bearer {auth_info['auth_token']}"}


def _get_login_response(client_id, attempt=0):
    if attempt == AUTH_TIMEOUT_SECONDS / 2:
        raise AuthTimeoutException()

    sleep(2)
    response = requests.get(f"{_BURLA_BACKEND_URL}/v1/login/{client_id}/token")

    if response.status_code == 404:
        return _get_login_response(client_id, attempt=attempt + 1)
    else:
        response.raise_for_status()
        return response.json()["token"], response.json()["email"]


def login():
    client_id = uuid4().hex
    login_url = f"{_BURLA_BACKEND_URL}/v1/login/{client_id}"

    if IN_COLAB:
        print(f"Please navigate to the following URL to login:\n\n    {login_url}\n")
        print(f"(We are unable to automatically open this from a Google Colab notebook)")
    else:
        print(f"Your browser has been opened to visit:\n\n    {login_url}\n")
        webbrowser.open(login_url)
    auth_token, email = _get_login_response(client_id)

    message = f"Thank you for registering with Burla! You are now logged in as [{email}].\n"
    message += "Please email jake@burla.dev with any questions!\n"
    print(message)

    if not CONFIG_PATH.exists():
        CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
        CONFIG_PATH.touch()
    CONFIG_PATH.write_text(json.dumps({"auth_token": auth_token, "email": email}))
